from typing import Optional


class FhirSenderException(Exception):
    def __init__(
        self,
        exception: Exception,
        url: str,
        json_data: str,
        response_text: Optional[str],
        response_status_code: Optional[int],
        message: str,
    ) -> None:
        self.exception: Exception = exception
        self.url: str = url
        self.data: str = json_data
        super().__init__(
            f"FHIR send failed to {url} {response_status_code}: {json_data}.  {message} {response_text}"
        )



import logging
import os
import shutil

from typing import Any, Generator

import boto3
import pytest
from _pytest.fixtures import FixtureFunctionMarker
from botocore.client import BaseClient
from moto import mock_ssm, mock_s3  # type: ignore
from pyspark.sql import SparkSession

# make sure env variables are set correctly
if "SPARK_HOME" not in os.environ:
    os.environ["SPARK_HOME"] = "/usr/local/opt/spark"


def quiet_py4j() -> None:
    """turn down spark logging for the carriers context"""
    logger = logging.getLogger("py4j")
    logger.setLevel(logging.ERROR)


def clean_spark_dir() -> None:
    """

    :return:
    """
    try:
        os.remove("./derby.log")
        shutil.rmtree("./metastore_db")
        shutil.rmtree("./spark-warehouse")
    except OSError:
        pass


def clean_spark_session(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    tables = session.catalog.listTables("default")

    for table in tables:
        print(f"clear_tables() is dropping table/view: {table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP TABLE IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS {table.name}")

    session.catalog.clearCache()


def clean_close(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    clean_spark_session(session)
    clean_spark_dir()
    session.stop()


@pytest.fixture(scope="session")
def spark_session(request: Any) -> SparkSession:
    # make sure env variables are set correctly
    if "SPARK_HOME" not in os.environ:
        os.environ["SPARK_HOME"] = "/usr/local/opt/spark"

    clean_spark_dir()

    master = "local[2]"

    jars = [
        "mysql:mysql-connector-java:8.0.24",
        "org.apache.spark:spark-sql-kafka-0-10_2.12:3.3.0",
        "io.delta:delta-core_2.12:2.1.0",
        "com.johnsnowlabs.nlp:spark-nlp_2.12:4.2.2",
        "org.apache.spark:spark-hadoop-cloud_2.12:3.3.1",
        "com.amazonaws:aws-java-sdk-bundle:1.12.339",
        "com.databricks:spark-xml_2.12:0.15.0",
    ]

    session = (
        SparkSession.builder.appName("pytest-pyspark-local-testing")
        .master(master)
        .config("spark.ui.showConsoleProgress", "false")
        .config("spark.sql.shuffle.partitions", "2")
        .config("spark.default.parallelism", "4")
        .config("spark.sql.broadcastTimeout", "2400")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config(
            "spark.sql.catalog.spark_catalog",
            "org.apache.spark.sql.delta.catalog.DeltaCatalog",
        )
        .config("spark.jars.packages", ",".join(jars))
        .enableHiveSupport()
        .getOrCreate()
    )

    request.addfinalizer(lambda: clean_close(session))
    quiet_py4j()
    return session


@pytest.fixture(scope="function")
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_REGION"] = "us-east-1"


@pytest.fixture(scope="function")
def ssm_mock(
    aws_credentials: FixtureFunctionMarker,
) -> Generator[BaseClient, None, None]:
    with mock_ssm():
        yield boto3.client("ssm", region_name="us-east-1")


@pytest.fixture(scope="function")
def s3_mock(
    aws_credentials: FixtureFunctionMarker,
) -> Generator[BaseClient, None, None]:
    with mock_s3():
        yield boto3.client("s3", region_name="us-east-1")



from typing import Dict, Any, Optional, List

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkDropViewsTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        views: List[str],
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.views: Param[List[str]] = Param(self, "views", "")
        self._setDefault(views=views)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        views: List[str] = self.getViews()

        view: str
        for view in views:
            if df.sparkSession.catalog.tableExists(tableName=view):
                self.logger.info(f"Dropping view {view}")
                df.sparkSession.catalog.dropTempView(viewName=view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getViews(self) -> List[str]:
        return self.getOrDefault(self.views)



from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, ArrayType
from spark_fhir_schemas.r4.complex_types.humanname import HumanNameSchema
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_bundle_splitter.v1.fhir_bundle_splitter import (
    FhirBundleSplitter,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_fhir_bundle_splitter_specified_resource_types(
    spark_session: SparkSession,
) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    df = df.sql_ctx.read.option("multiLine", True).json(
        str(data_dir.joinpath("test.json"))
    )
    df.createOrReplaceTempView("fhir_bundle")

    # Act
    with ProgressLogger() as progress_logger:
        FhirBundleSplitter(
            view="fhir_bundle",
            resource_types=[
                "Practitioner",
                "PractitionerRole",
                "Organization",
                "Location",
                "HealthcareService",
            ],
            progress_logger=progress_logger,
            temp_folder=temp_folder,
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("fhir_bundle")
    result_df.show(truncate=False)
    result_df.printSchema()
    result_df.select("practitioner").printSchema()

    assert "practitioner" in result_df.columns
    # noinspection SpellCheckingInspection
    assert "practitionerrole" in result_df.columns
    assert "location" in result_df.columns

    name_schema = result_df.select("practitioner.name").schema
    assert str(name_schema) == str(
        StructType([StructField("name", ArrayType(HumanNameSchema.get_schema()))])
    )



import json
from typing import Any, Callable, Dict, List, Optional, Union, Generator

from spark_pipeline_framework.utilities.api_helper.http_request import HelixHttpRequest
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import sc


class HttpDataReceiver(FrameworkTransformer):
    """
    This is a generic class to call a http api and return the response
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        http_request_generator: Callable[
            [DataFrame, Optional[ProgressLogger]],
            Generator[HelixHttpRequest, None, None],
        ],
        view_name: str,
        name: Optional[str] = None,
        response_processor: Optional[
            Callable[
                [
                    List[Dict[str, Any]],
                    Union[List[Dict[str, Any]], Dict[str, Any]],
                    HelixHttpRequest,
                    Optional[ProgressLogger],
                ],
                List[Dict[str, Any]],
            ]
        ] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        log_response: bool = False,
    ) -> None:
        """
        Transformer to call and receive data from an API


        :param http_request_generator: Generator to build next http request
        :param view_name: name of the view to read the response into
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        it supposed to return a HelixHttpRequest to be used to call the API or return None to end the API call loop
        :param response_processor: it can change the result before loading to spark df
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.http_request_generator: Param[
            Callable[
                [DataFrame, Optional[ProgressLogger]],
                Generator[HelixHttpRequest, None, None],
            ]
        ] = Param(self, "http_request_generator", "")
        self._setDefault(http_request_generator=None)

        self.view_name: Param[str] = Param(self, "view_name", "")
        self._setDefault(view_name=None)

        self.log_response: Param[bool] = Param(self, "log_response", "")
        self._setDefault(log_response=log_response)

        self.response_processor: Param[
            Optional[
                Callable[
                    [
                        List[Dict[str, Any]],
                        Dict[str, Any],
                        HelixHttpRequest,
                        Optional[ProgressLogger],
                    ],
                    List[Dict[str, Any]],
                ]
            ]
        ] = Param(self, "response_processor", "")
        self._setDefault(response_processor=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        http_request_generator: Callable[
            [DataFrame, Optional[ProgressLogger]],
            Generator[HelixHttpRequest, None, None],
        ] = self.getHttpRequestGenerator()
        view_name = self.getView()
        name: Optional[str] = self.getName()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=f"{name}_http_data_receiver", progress_logger=progress_logger
        ):
            responses: List[Dict[str, Any]] = []
            try:
                for http_request in http_request_generator(df, progress_logger):
                    self.logger.debug(f"Calling API: {http_request.to_string()}...")
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Calling API: {http_request.to_string()}..."
                        )
                    response = http_request.get_result()

                    self.logger.info(
                        f"Successfully retrieved: {http_request.url} with status {response.status}"
                    )
                    if self.getLogResponse():
                        self.logger.info(
                            f"Response: {json.dumps(response.result, default=str)}"
                        )
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Successfully retrieved: {http_request.url} with status {response.status}"
                        )
                        if self.getLogResponse():
                            progress_logger.write_to_log(
                                f"Response [{response.status}]: {json.dumps(response.result, default=str)}"
                            )

                    # accumulated responses before loading to spark
                    response_processor = self.getResponseProcessor()
                    if response_processor:
                        responses = response_processor(
                            responses, response.result, http_request, progress_logger
                        )
                    else:
                        responses.append(response.result)

                df2 = df.sql_ctx.read.json(
                    sc(df).parallelize([json.dumps(r) for r in responses])
                )
                df2.createOrReplaceTempView(view_name)
            except Exception as e:
                url = (
                    http_request.url
                    if (http_request and hasattr(http_request, "url"))
                    else None
                )
                error_text: str = f"Url: {url}: {http_request} {repr(e)}"
                self.logger.error(error_text)
                raise Exception(error_text) from e

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHttpRequestGenerator(
        self,
    ) -> Callable[
        [DataFrame, Optional[ProgressLogger]], Generator[HelixHttpRequest, None, None]
    ]:
        return self.getOrDefault(self.http_request_generator)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view_name)

    # noinspection PyPep8Naming
    def getResponseProcessor(
        self,
    ) -> Optional[
        Callable[
            [
                List[Dict[str, Any]],
                Dict[str, Any],
                HelixHttpRequest,
                Optional[ProgressLogger],
            ],
            List[Dict[str, Any]],
        ]
    ]:
        return self.getOrDefault(self.response_processor)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogResponse(self) -> bool:
        return self.getOrDefault(self.log_response)



import json
from typing import Any, Callable, Dict, List, Optional

from spark_pipeline_framework.transformers.athena_table_creator.v1.athena_table_creator import (
    AthenaTableCreator,
)
from spark_pipeline_framework.transformers.fhir_exporter.v1.fhir_exporter import (
    FhirExporter,
)
from spark_pipeline_framework.transformers.fhir_sender.v1.fhir_sender import FhirSender
from spark_pipeline_framework.transformers.send_automapper_to_fhir.exceptions.automapper_to_fhir_transformer_exception import (
    AutomapperToFhirTransformerException,
)
from spark_pipeline_framework.utilities.athena.athena_source_file_type import (
    AthenaSourceFileType,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.proxy_base import ProxyBase
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class AutoMapperToFhirTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        transformer: ProxyBase,
        func_get_path: Callable[[str, str], str],
        func_get_response_path: Callable[[str, str], str],
        fhir_server_url: str,
        source_entity_name: str,
        fhir_validation_url: Optional[str] = None,
        athena_schema: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        send_to_fhir: Optional[bool] = True,
    ):
        """
        Runs the auto-mappers, saves the result to Athena db and then sends the results to fhir server

        :param transformer: transformer to run
        :param func_get_path: function that gets the local path for a resource
        :param func_get_response_path: function that gets the path to save the responses from the FHIR server
        :param fhir_server_url: url to FHIR server
        :param fhir_validation_url: url to FHIR validation server
        :param source_entity_name: resource name
        :param athena_schema: schema to use when writing data to Athena
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert fhir_server_url

        self.logger = get_logger(__name__)
        self.logger.info(
            f"Parameters={json.dumps({str(key): str(value) for key, value in parameters.items()}) if parameters else 'None'}"
        )

        # add a param
        self.transformer: Param[ProxyBase] = Param(self, "transformer", "")
        self._setDefault(transformer=transformer)

        self.func_get_path: Param[Callable[[str, str], str]] = Param(
            self, "func_get_path", ""
        )
        self._setDefault(func_get_path=func_get_path)

        self.func_get_response_path: Param[Callable[[str, str], str]] = Param(
            self, "func_get_response_path", ""
        )
        self._setDefault(func_get_response_path=func_get_response_path)

        self.fhir_server_url: Param[str] = Param(self, "fhir_server_url", "")
        self._setDefault(fhir_server_url=fhir_server_url)

        self.fhir_validation_url: Param[Optional[str]] = Param(
            self, "fhir_validation_url", ""
        )
        self._setDefault(fhir_validation_url=fhir_validation_url)

        self.source_entity_name: Param[str] = Param(self, "source_entity_name", "")
        self._setDefault(source_entity_name=source_entity_name)

        self.athena_schema: Param[Optional[str]] = Param(self, "athena_schema", "")
        self._setDefault(athena_schema=athena_schema)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[Optional[List[str]]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.send_to_fhir: Param[Optional[bool]] = Param(self, "send_to_fhir", "")
        self._setDefault(send_to_fhir=send_to_fhir)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        transformer: ProxyBase = self.getTransformer()
        func_get_path: Callable[[str, str], str] = self.getFuncGetPath()
        func_get_response_path: Callable[
            [str, str], str
        ] = self.getFuncGetResponsePath()
        fhir_server_url: str = self.getFhirServerUrl()
        fhir_validation_url: Optional[str] = self.getFhirValidationServerUrl()
        name: Optional[str] = self.getName()
        source_entity_name: str = self.getSourceEntityName()
        athena_schema: Optional[str] = self.getAthenaSchema()
        parameters = self.getParameters()
        assert parameters
        progress_logger = self.getProgressLogger()

        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()
        send_to_fhir: Optional[bool] = self.getSendToFhir()

        self.logger.info(f"Calling {fhir_server_url} with client_id={auth_client_id}")

        if len(transformer.transformers) == 0:
            raise AutomapperToFhirTransformerException(
                f"No transformers found for '{transformer.name}'."
            )

        t: Transformer
        for t in transformer.transformers:
            if isinstance(t, FrameworkMappingLoader):
                t.transform(df)  # run the automapper
                views: List[str] = t.getViews()
                view: str
                for view in views:
                    self.logger.info(
                        f"---- Started processing for view: {view} --------"
                    )
                    # get resource name
                    result_df: DataFrame = df.sql_ctx.table(view)
                    if spark_is_data_frame_empty(df=result_df):
                        self.logger.info(f"No data to export/send for view '{view}'")
                        continue
                    self.logger.info(f"----- view: {view} has rows-------")
                    first_row = result_df.select("resourceType").first()
                    assert first_row
                    # noinspection PyPep8Naming
                    resourceType: str = first_row["resourceType"]
                    fhir_resource_path: str = func_get_path(view, resourceType)
                    fhir_resource_response_path: str = func_get_response_path(
                        view, resourceType
                    )
                    # export as FHIR to local fhir_resource_path
                    FhirExporter(
                        view=view,
                        file_path=fhir_resource_path,
                        parameters=parameters,
                        progress_logger=progress_logger,
                        name=f"{name}_fhir_exporter",
                    ).transform(df)
                    if fhir_resource_path.startswith("s3") and athena_schema:
                        # create table in Athena
                        s3_temp_folder: str = str(parameters.get("s3_temp_folder"))
                        AthenaTableCreator(
                            view=view,
                            schema_name=athena_schema,
                            table_name=f"{source_entity_name}_{resourceType}",
                            s3_source_path=fhir_resource_path,
                            source_file_type=AthenaSourceFileType.JSONL,
                            s3_temp_folder=s3_temp_folder,
                            parameters=parameters,
                            progress_logger=progress_logger,
                        ).transform(df)
                    # send to FHIR server
                    if send_to_fhir:
                        FhirSender(
                            resource=resourceType,
                            server_url=fhir_server_url,
                            validation_server_url=fhir_validation_url,
                            file_path=fhir_resource_path,
                            response_path=fhir_resource_response_path,
                            parameters=parameters,
                            progress_logger=progress_logger,
                            batch_size=parameters.get("batch_size", 0),
                            throw_exception_on_validation_failure=parameters.get(
                                "throw_exception_on_validation_failure"
                            ),
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            name=f"{name}_fhir_sender",
                        ).transform(df)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTransformer(self) -> ProxyBase:
        return self.getOrDefault(self.transformer)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetPath(self) -> Callable[[str, str], str]:
        return self.getOrDefault(self.func_get_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetResponsePath(self) -> Callable[[str, str], str]:
        return self.getOrDefault(self.func_get_response_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFhirServerUrl(self) -> str:
        return self.getOrDefault(self.fhir_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFhirValidationServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.fhir_validation_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSourceEntityName(self) -> str:
        return self.getOrDefault(self.source_entity_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAthenaSchema(self) -> Optional[str]:
        return self.getOrDefault(self.athena_schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSendToFhir(self) -> Optional[bool]:
        return self.getOrDefault(self.send_to_fhir)



from logging import Logger
from typing import List, Optional, Dict, Any

from helix_fhir_client_sdk.fhir_client import FhirClient
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse

from spark_pipeline_framework.utilities.fhir_helpers.get_fhir_client import (
    get_fhir_client,
)


def send_json_bundle_to_fhir(
    json_data_list: List[str],
    server_url: str,
    validation_server_url: Optional[str],
    resource: str,
    logger: Logger,
    auth_server_url: Optional[str],
    auth_client_id: Optional[str],
    auth_client_secret: Optional[str],
    auth_login_token: Optional[str],
    auth_scopes: Optional[List[str]],
    auth_access_token: Optional[str],
    log_level: Optional[str],
    retry_count: Optional[int] = None,
    exclude_status_codes_from_retry: Optional[List[int]] = None,
) -> Optional[FhirMergeResponse]:
    fhir_client: FhirClient = get_fhir_client(
        logger=logger,
        server_url=server_url,
        auth_server_url=auth_server_url,
        auth_client_id=auth_client_id,
        auth_client_secret=auth_client_secret,
        auth_login_token=auth_login_token,
        auth_access_token=auth_access_token,
        auth_scopes=auth_scopes,
        log_level=log_level,
    )

    fhir_client = fhir_client.resource(resource)
    if validation_server_url:
        fhir_client = fhir_client.validation_server_url(validation_server_url)

    if retry_count is not None:
        fhir_client = fhir_client.retry_count(retry_count)

    if exclude_status_codes_from_retry:
        fhir_client = fhir_client.exclude_status_codes_from_retry(
            exclude_status_codes_from_retry
        )

    try:
        logger.debug("----------- Sending data to FHIR -------")
        logger.debug(json_data_list)
        logger.debug("----------- End sending data to FHIR -------")
        response: FhirMergeResponse = fhir_client.merge(json_data_list)
        return response
    except AssertionError as e:
        logger.exception(
            Exception(
                f"Assertion: FHIR send failed: {str(e)} for resource: {json_data_list}"
            )
        )
        return None


def send_fhir_delete(
    obj_id: str,
    server_url: str,
    resource: str,
    logger: Logger,
    auth_server_url: Optional[str],
    auth_client_id: Optional[str],
    auth_client_secret: Optional[str],
    auth_login_token: Optional[str],
    auth_scopes: Optional[List[str]],
    auth_access_token: Optional[str],
    log_level: Optional[str],
) -> Dict[str, Any]:
    fhir_client: FhirClient = get_fhir_client(
        logger=logger,
        server_url=server_url,
        auth_server_url=auth_server_url,
        auth_client_id=auth_client_id,
        auth_client_secret=auth_client_secret,
        auth_login_token=auth_login_token,
        auth_access_token=auth_access_token,
        auth_scopes=auth_scopes,
        log_level=log_level,
    )

    fhir_client = fhir_client.resource(resource)
    fhir_client.resource(resource).id_(obj_id)
    try:
        response: FhirDeleteResponse = fhir_client.delete()
        if response and response.status == 200:
            return {"deleted": True}
        else:
            return {"deleted": False, "issue": f"Failed to delete {resource}: {obj_id}"}
    except AssertionError as e:
        logger.exception(
            Exception(f"Assertion: FHIR send failed: {str(e)} for {resource}: {obj_id}")
        )
        return {"issue": str(e)}
